import os
import functools
from typing import Dict, Optional, Any, List, Tuple
import json
import requests

from . import exceptions

RESPONSE_EXCEPTIONS = {
    400: exceptions.BadRequest,
    500: exceptions.PPAServerError,
}


def event(name: str, text: str, files: Optional[List[Tuple]] = None, **kwargs):
    if files:
        return requests.post(f"{os.environ['PPA_URL']}/v1/{name}", dict(text=text, **kwargs), files=files)
    return requests.post(f"{os.environ['PPA_URL']}/v1/{name}", json=dict(text=text, **kwargs))


def task_event(status_code_exceptions: Optional[Dict[int, Exception]] = None, return_tuple=None):
    def _catch_status_code(status_code: int):
        try:
            raise status_code_exceptions[status_code]
        except (TypeError, KeyError):
            pass

    def _wrapped_task_event(func):
        @functools.wraps(func)
        def _process_task_event(*args, **kwargs) -> Any:
            def _send_request(request_func) -> requests.Response:
                try:
                    response = request_func()
                    response.raise_for_status()
                    return response
                except requests.exceptions.HTTPError as e:
                    # If the function has been decorated with custom exceptions then we may be able to match one
                    # of the status codes here. Note: not every non-200 status code will raise a urllib exception.
                    _catch_status_code(e.response.status_code)
                    try:
                        # Nothing provided in the decorator has caught the issue yet, so raise
                        # generic 400 and 500 exceptions if appropriate.
                        raise RESPONSE_EXCEPTIONS[e.response.status_code](str(e))
                    except KeyError:
                        # The request exception isn't handled and will be re-raised below.
                        pass
                    raise e

            response = _send_request(functools.partial(func, *args, **kwargs))
            _catch_status_code(response.status_code)
            try:
                raise RESPONSE_EXCEPTIONS[response.status_code](response.text)
            except KeyError:
                pass
            # Any response that made it this far has been deemed acceptable. We don't care
            # about the status code anymore, just return response_json (which can be None
            # if the event is a UI output).
            try:
                response_json = response.json()
            except json.decoder.JSONDecodeError:
                return None

            return return_tuple(**response_json) if return_tuple else response_json

        return _process_task_event

    return _wrapped_task_event
